-- -*- mode: Lua; indent-tabs-mode: nil; lua-indent-level: 4; -*-

--[[
    
    Headshots v1.0
    by irons
    
]]

--[[ EASY TWEAKS BELOW.  defaults derived from marathon infinity ]]

-- ! the amount of extra damage that a headshot causes.  at present,
-- ! the amount of damage each projectile deals is not visible to lua
-- ! scripts, so this can't be a percentage (unless you add your own
-- ! lookup).  for reference, a player at "red" or "1x" shields has
-- ! 150 points.
-- !
-- ! set to zero or below to cause instant death regardless of the
-- ! monster's starting health.
-- !
-- ! set to a value between zero and one to make this a percentage of
-- ! the monster's base health.  for example, 0.75 would cause all
-- ! monsters to receive damage equal to 75% of their base health.  if
-- ! you want to deal 100% (i.e. probably instantly kill), see above
-- ! about setting this value to zero or below.
-- !
-- ! it is possible to override this per monster type in the
-- ! #HEADS_BY_MONSTER_TYPE table.
HEADSHOT_BONUS_DAMAGE = 0

-- ! the list of monster types that *do not* receive extra damage from
-- ! a headshot.  the monster type names are in the `MonsterTypes`
-- ! section of the lua documentation.
EXCLUDED_VICTIM_MONSTERS_BY_TYPE = {
    -- ticks are all (or no) head.  one-shot kill anyway
    "minor tick",
    "major tick",
    -- drones don't really have heads
    "minor drone",
    "major drone",
    "big minor drone",
    "big major drone",
    "possessed drone",
    -- juggernaut head is just a hood ornament afaik.  also,
    -- juggernaut head is at a horizontal offset from its center.
    -- without being able to account for hit angle, this would allow
    -- hits on its "side" to register as headshots.
    "minor juggernaut",
    "major juggernaut",
}

-- ! the list of projectile types that *do* give extra damage via
-- ! headshot, assuming the victim monster is not of a type in the
-- ! table #EXCLUDED_VICTIM_MONSTERS_BY_TYPE and the aggressor monster
-- ! is in the table #INCLUDED_AGGRESSOR_MONSTERS_BY_TYPE.  the
-- ! projectile type names are in the `ProjectileTypes` section of the
-- ! lua documentation.
INCLUDED_PROJECTILES_BY_TYPE = {
    "pistol bullet", -- note: used in both pistol and SMG
    "fusion bolt major",
}

-- ! the list of monster types whose projectiles *do* give extra
-- ! damage via headshot, assuming the victim monster is not of a type
-- ! in the table #EXCLUDED_VICTIM_MONSTERS_BY_TYPE and the projectile
-- ! that hit the monster is in #INCLUDED_PROJECTILES_BY_TYPE.  the
-- ! monster type names are in the `MonsterTypes` section of the lua
-- ! documentation.
INCLUDED_AGGRESSOR_MONSTERS_BY_TYPE = {
    "player",
    "security bob",
}

-- ! the percentage of a monster between its lowest point and where
-- ! its head starts.  everything above this counts as the head.
-- ! might be weird on juggernauts and similar monsters.  see below
-- ! for a solution.
DEFAULT_HEAD_HEIGHT_PERCENT = 0.89

-- ! the sound to play for a player who just made a headshot.  set to
-- ! nil or empty string if you want no sound.  sounds are in the
-- ! `Sounds` section of the documentation.
PLAYER_SHOOTER_SOUND = "puzzle switch"

-- ! the fader to play for a player who just made a headshot.  set to
-- ! nil or empty string if you want no fader.  faders are in the
-- ! `FaderTypes` section of the documentation.
PLAYER_SHOOTER_FADE = "dodge yellow"

-- ! the sound to play for a player whose head was just shot.  set to
-- ! nil or empty string if you want no sound.  sounds are in the
-- ! `Sounds` section of the documentation.
PLAYER_VICTIM_SOUND = "empty gun"

-- ! the fader to play for a player whose head was just shot.  set to
-- ! nil or empty string if you want no fader.  faders are in the
-- ! `FaderTypes` section of the documentation.
PLAYER_VICTIM_FADE = "dodge purple"

-- ! a table listing the lowest/highest point of a monster's head vs.
-- ! the monster's lowest point by monster type.  you can store this
-- ! in absolute world units (`min_absolute`/`max_absolute`) or you
-- ! can store it in percentages (`min_percent`/`max_percent`).
-- !
-- ! if you do not populate a max value, the script assumes that
-- ! everything higher than the head start includes the head.  if you
-- ! want the head to start at the monster's lowest point, set the min
-- ! to 0; the max determines where the top of the head is.
-- !
-- ! you can override the value of #HEADSHOT_BONUS_DAMAGE per monster
-- ! type by populating `damage` in following similar guidelines.
-- !
-- ! i have included some sample entries below.  the monster type
-- ! names are in the `MonsterTypes` section of the lua documentation.
HEADS_BY_MONSTER_TYPE = {
    -- player is harder to hit
    ["player"] = { min_absolute = 0.85, damage = 100 },
    -- troopers have helmets
    ["minor trooper"] = { damage = 0.50 },
    ["major trooper"] = { damage = 0.50 },
    -- compilers have metal heads
    ["minor compiler"] = { damage = 0.40 },
    ["major compiler"] = { damage = 0.40 },
    ["minor invisible compiler"] = { damage = 0.40 },
    ["major invisible compiler"] = { damage = 0.40 },
    -- fuck these guys
    ["minor enforcer"] = { min_percent = 0.80, damage = 1000 },
    ["major enforcer"] = { min_percent = 0.80, damage = 1000 },
    -- hunters have thick skulls(?)
    ["minor hunter" ] = { damage = 0.45 },
    ["major hunter" ] = { damage = 0.45 },
}

--[[ NO CHANGES AFTER THIS POINT UNLESS YOU KNOW WHAT YOU ARE DOING ]]


Triggers = {}

-- the default top-of-head value is the top of the whole
-- monster
DEFAULT_MAX_PERCENT = 1.0

-- NOTE: can't really do this with current functionality.  leaving
-- in to save effort if this changes.
--
-- the full width of every monster's head as a percentage of its
-- diameter.
DEFAULT_HEAD_WIDTH_PERCENT = 0.10

DEFAULT_ENTRY = {
    min_percent = DEFAULT_HEAD_HEIGHT_PERCENT,
    width_percent = DEFAULT_HEAD_WIDTH_PERCENT,
    max_percent = DEFAULT_MAX_PERCENT,
    damage = HEADSHOT_BONUS_DAMAGE,
}

function Triggers.init()
    handle_invalid_entries()
    precalculate_absolutes()
    convert_indexes()
    set_reasonable_defaults()
end

function Triggers.projectile_detonated(p_type, owner, poly, x, y, z, t)
    detect_headshot(p_type, owner, poly, x, y, z, t)
end

function detect_headshot(p_type, owner, poly, x, y, z, t)
    if not owner or not is_monster(t) or not t.valid then
        return
    end

    p_type = tostring(p_type)
    local t_type = tostring(t.type)
    local o_type = tostring(owner.type)
    
    if INCLUDED_PROJECTILES_BY_TYPE[p_type]
        and owner ~= t
        and INCLUDED_AGGRESSOR_MONSTERS_BY_TYPE[o_type]
        and not EXCLUDED_VICTIM_MONSTERS_BY_TYPE[t_type]
    then
        local entry = HEADS_BY_MONSTER_TYPE[t_type]
        
        if z >= t.z + entry.min_absolute
            and z <= t.z + entry.max_absolute
        then
            -- there doesn't seem to be a way to do head
            -- width: projectile_detonated does not let us
            -- know which way the projectile was facing.
            -- without that, we can't know if the hit was
            -- "glancing" or "dead center": from the
            -- perspective of a point on a circle, there is no
            -- such thing without a direction
            
            if entry.damage <= 0 then
                t.life = 0
            else
                t.life = t.life - entry.damage
            end

            local killed

            if t.player then
                killed = p.player.index
                if PLAYER_VICTIM_SOUND then
                    owner.player:play_sound(PLAYER_VICTIM_SOUND)
                end
                if PLAYER_VICTIM_FADE then
                    owner.player:fade_screen(PLAYER_VICTIM_FADE)
                end
            end
            if owner.player then
                if t.life == 0 then
                    t:damage(100, "energy drain")
                    if killed then
                        -- make up score for sourceless
                        -- damage/owner when killing outright
                        owner.player.kills[killed] =
                            owner.player.kills[killed] + 1
                    end
                end
                if PLAYER_SHOOTER_SOUND then
                    owner.player:play_sound(PLAYER_SHOOTER_SOUND)
                end
                if PLAYER_SHOOTER_FADE then
                    owner.player:fade_screen(PLAYER_SHOOTER_FADE)
                end
            end
        end
    end
end

function handle_invalid_entries()
    local k, v, tmp
    
     -- fix invalid head spec entries
    for k, v in pairs(HEADS_BY_MONSTER_TYPE) do
	if not (v.min_percent or v.min_absolute) then
            v.min_percent = DEFAULT_HEAD_HEIGHT_PERCENT
        end
	if not (v.width_percent or v.width_absolute) then
		v.width_percent = DEFAULT_HEAD_WIDTH_PERCENT
	    end
        if not (v.max_percent or v.max_absolute) then
		v.max_percent = DEFAULT_MAX_PERCENT
	    end
    end
    
    -- clean up after iteration
    for k, v in MonsterTypes() do
	k = tostring(k)
	if not HEADS_BY_MONSTER_TYPE[k] then
	    HEADS_BY_MONSTER_TYPE[k] = copy_default_entry()
	end
    end
end

function precalculate_absolutes()
    local k, v, m_type
    
    -- precalculate absolute heights/widths for percentages
    for k, v in pairs(HEADS_BY_MONSTER_TYPE) do
	m_type = MonsterTypes[k]
	if m_type then
	    if not v.min_absolute then
		v.min_absolute = v.min_percent * m_type.height
	    end
	    if not v.max_absolute then
		v.max_absolute = v.max_percent * m_type.height
	    end
	    if not v.width_absolute then
		-- simple "width" definition = diameter
		v.width_absolute = v.width_percent * m_type.radius * 2
	    end
            if not v.damage then
                v.damage = HEADSHOT_BONUS_DAMAGE
            end
            if v.damage > 0 and v.damage < 1 then
                local vitality = m_type.vitality
                if m_type == "player" then
                    vitality = 150
                elseif m_type.alien then
                    vitality = adjust_vitality_for_difficulty(vitality)
                end
                v.damage = math.floor(vitality * v.damage)
            end
	end
    end
end

function convert_indexes()
    local x, k, v
    
    -- convert integer-based tables to string-based
    local index_tables = {
	INCLUDED_PROJECTILES_BY_TYPE,
	INCLUDED_AGGRESSOR_MONSTERS_BY_TYPE,
	EXCLUDED_VICTIM_MONSTERS_BY_TYPE,
    }

    for _, v in ipairs(index_tables) do
	for x, k in ipairs(v) do
	    v[x] = nil
	    v[k] = true
	end
    end
end

function set_reasonable_defaults()
    if PLAYER_SHOOTER_SOUND == "" then
	PLAYER_SHOOTER_SOUND = nil
    end
    if PLAYER_SHOOTER_SOUND and not Sounds[PLAYER_SHOOTER_SOUND] then
	PLAYER_SHOOTER_SOUND = nil
    end
    if PLAYER_VICTIM_SOUND == "" then
	PLAYER_VICTIM_SOUND = nil
    end
    if PLAYER_VICTIM_SOUND and not Sounds[PLAYER_VICTIM_SOUND] then
	PLAYER_VICTIM_SOUND = nul
    end
    if PLAYER_SHOOTER_FADE == "" then
	PLAYER_SHOOTER_FADE = nil
    end
    if PLAYER_SHOOTER_FADE and not FadeTypes[PLAYER_SHOOTER_FADE] then
	PLAYER_SHOOTER_FADE = nil
    end
    if PLAYER_VICTIM_FADE == "" then
	PLAYER_VICTIM_FADE = nil
    end
    if PLAYER_VICTIM_FADE and not FadeTypes[PLAYER_VICTIM_FADE] then
	PLAYER_VICTIM_FADE = nil
    end
end

function copy_default_entry()
    return {
        min_percent = DEFAULT_ENTRY.min_percent,
        max_percent = DEFAULT_ENTRY.max_percent,
        min_absolute = DEFAULT_ENTRY.min_absolute,
        max_absolute = DEFAULT_ENTRY.max_absolute,
        width_percent = DEFAULT_ENTRY.width_percent,
        width_absolute = DEFAULT_ENTRY.width_absolute,
        damage = DEFAULT_ENTRY.damage,
    }
end

function adjust_vitality_for_difficulty(vitality)
    -- aleph one says:
    -- kindergarten:  - (vitality >> 1)
    -- easy:          - (vitality >> 2)
    -- major damage:  + (vitality >> 2)
    -- total carnage: + (vitality >> 1)
    -- for our purposes, right shift is the same as divide by 2.
    -- these are integers, so this always rounds down.

    if Game.difficulty == "kindergarten" then
        return vitality - math.floor(vitality / 2)
    elseif Game.difficulty == "easy" then
        return vitality - math.floor(vitality / 4)
    elseif Game.difficulty == "major damage" then
        return vitality + math.floor(vitality / 4)
    elseif Game.difficulty == "total carnage" then
        return vitality + math.floor(vitality / 2)
    else
        return vitality
    end
end
